#=======================================================================#
# SCRIPT : gridplus.tcl                                                 #
# PURPOSE: Gridplus layout manager.                                     #
# AUTHOR : Adrian Davis                                                 #
# VERSION: 1.2                                                          #
# DATE   : 12/04/2005                                                   #
#=======================================================================#

package require Tk 8.4

package require msgcat
namespace import msgcat::*

catch {package require icons}
catch {package require tablelist}

package provide gridplus 1.2

#=======================================================================#
# Export the public interface.                                          #
#=======================================================================#

namespace eval ::gridplus:: {
   namespace export gridplus
   namespace export gpmap
   namespace export gpset
   namespace export gpunset
   variable gpButtonHelp
   variable gpConfig
   variable gpDateFormat
   variable gpFocus
   variable gpGroup
   variable gpGroupState
   variable gpInfo
   variable gpStyles
   variable gpTabOrder
   variable gpValidations
}

#=======================================================================#
# PROC   : ::gridplus::gridplus                                         #
# PURPOSE: Exported command.                                            #
#=======================================================================#

proc ::gridplus::gridplus {args} {
   variable gpConfig
   variable gpDateFormat

   # If first call run initialisation.
   if {! [info exists gpConfig]} {
      gpInit
   }

   # Set array of valid/default options.
   array set options [list                                                            \
      -action           none                                                          \
      -background       [gpGetOption Gridplus.background [. cget -background]]        \
      -borderwidth      [gpGetOption Gridplus.borderWidth 2]                          \
      -buttoncolor      [gpGetOption Gridplus.buttonColor black/[. cget -background]] \
      -buttonrelief     [gpGetOption Gridplus.buttonRelief raised]                    \
      -century          $gpDateFormat(century)                                        \
      -columnsort       1                                                             \
      -compound         left                                                          \
      -dateformat       $gpConfig(dateformat)                                         \
      -deletewindow     {}                                                            \
      -errormessage     $gpConfig(errormessage)                                       \
      -fileicon         [gpGetOption Gridplus.fileIcon file]                          \
      -foldericon       [gpGetOption Gridplus.folderIcon folder]                      \
      -foreground       [gpGetOption Gridplus.foreground black]                       \
      -group            {}                                                            \
      -height           10                                                            \
      -icon             [gpGetOption Gridplus.icon {}]                                \
      -iconfile         $gpConfig(iconfile)                                           \
      -iconpath         $gpConfig(iconpath)                                           \
      -insertexpr       {}                                                            \
      -insertoptions    {}                                                            \
      -irelief          [gpGetOption Gridplus.iRelief sunken]                         \
      -justify          left                                                          \
      -labelcolor       [gpGetOption Gridplus.labelColor black/black]                 \
      -labelstyle       [gpGetOption Gridplus.labelStyle /]                           \
      -linerelief       [gpGetOption Gridplus.lineRelief sunken]                      \
      -linewidth        [gpGetOption Gridplus.lineWidth 2]                            \
      -linkcolor        [gpGetOption Gridplus.linkColor black/black]                  \
      -linkcursor       [gpGetOption Gridplus.linkCursor arrow]                       \
      -linkstyle        [gpGetOption Gridplus.linkStyle /underline]                   \
      -listvariable     {}                                                            \
      -modal            0                                                             \
      -options          {}                                                            \
      -pad              [gpGetOption Gridplus.pad 5]                                  \
      -padx             [gpGetOption Gridplus.padX [gpGetOption Gridplus.pad 5]]      \
      -pady             [gpGetOption Gridplus.padY [gpGetOption Gridplus.pad 5]]      \
      -pattern          {}                                                            \
      -prefix           $gpConfig(prefix)                                             \
      -proc             $gpConfig(proc)                                               \
      -relief           [gpGetOption Gridplus.relief flat]                            \
      -scroll           none                                                          \
      -selectfirst      0                                                             \
      -selectmode       [gpGetOption Girdplus.selectMode browse]                      \
      -size             8                                                             \
      -sortfirst        0                                                             \
      -sortorder        increasing                                                    \
      -space            [gpGetOption Gridplus.space 20]                               \
      -state            normal                                                        \
      -style            {}                                                            \
      -subst            [gpGetOption Gridplus.subst 1]                                \
      -taborder         column                                                        \
      -takefocus        1                                                             \
      -tags             0                                                             \
      -text             {}                                                            \
      -title            {}                                                            \
      -treemode         [gpGetOption Gridplus.treeMode contract]                      \
      -validation       {}                                                            \
      -variables        1                                                             \
      -width            40                                                            \
      -wrap             word                                                          \
      -wraplength       0                                                             \
   ]

   # Read mode.
   set mode [lindex $args 0]

   # Validate mode and set parameter template.
   switch -- $mode {
      add         {set argTemplate [list "name 1" "options 2 end"]}
      button      {set argTemplate [list "name 1" "options 2 end-1" "layout end"]}
      checkbutton {set argTemplate [list "name 1" "options 2 end-1" "layout end"]}
      clear       {set argTemplate [list "name 1" "options 2 end"]}
      entry       {set argTemplate [list "name 1" "options 2 end-1" "layout end"]}
      goto        {set argTemplate [list "name 1" "options 2 end-1" "layout end"]}
      grid        {set argTemplate [list "name 1" "options 2 end-1" "layout end"]}
      init        {set argTemplate [list "options 1 end"]}
      layout      {set argTemplate [list "name 1" "options 2 end-1" "layout end"]}
      line        {set argTemplate [list "name 1" "options 2 end"]}
      link        {set argTemplate [list "name 1" "options 2 end-1" "layout end"]}
      menu        {set argTemplate [list "name 1" "options 2 end-1" "layout end"]}
      menubutton  {set argTemplate [list "name 1" "options 2 end-1" "layout end"]}
      notebook    {set argTemplate [list "name 1" "options 2 end-1" "layout end"]}
      radiobutton {set argTemplate [list "name 1" "options 2 end-1" "layout end"]}
      set         {set argTemplate [list "options 1 end"]}
      style       {set argTemplate [list "name 1" "options 2 end-1" "layout end"]}
      tablelist   {set argTemplate [list "name 1" "options 2 end-1" "layout end"]}
      text        {set argTemplate [list "name 1" "options 2 end"]}
      tree        {set argTemplate [list "name 1" "options 2 end"]}
      window      {set argTemplate [list "name 1" "options 2 end"]}
      default     {error "GRIDPLUS ERROR: Invalid mode ($mode)"}
   }

   # Check if sufficient args.
   if {[llength $args] < [llength $argTemplate]} {
      error "GRIDPLUS ERROR: Wrong number of Args"
   }

   # Check if sufficient args remain for option/value pairs.
   if {[expr {([llength $args] - [llength $argTemplate]) % 2}] != 0} {
      error "GRIDPLUS ERROR: Unmatched option/value"
   }

   # Read/validate arguments.
   foreach template $argTemplate {
      set argName  [lindex $template 0]
      set argStart [lindex $template 1]
      set argEnd   [lindex $template 2]
      # If argName is "options" read option/value pairs.
      if {$argName eq "options"} {
         foreach {option value} [lrange $args $argStart $argEnd] {
            if {[info exists options($option)]} {
               switch -- $option {
                  -pad {
                     set options(-padx) $value
                     set options(-pady) $value
                  }
                  -title {
                     set options(-title) $value
                     if {$options(-title) ne ""} {
                        set options(-relief) groove
                     }
                  }
                  default {
                     set options($option) $value
                  }
               }
            } else {
               error "GRIDPLUS ERROR: Invalid option ($option)"
            }
         }
      } else {
         set options($argName) [lindex $args $argStart]
      }
   }

   # Set style.
   gpSetStyle

   # Remove blank lines from "layout".
   if {[info exists options(layout)]} {
      regsub -all -- {\n\n}                $options(layout) "\n" options(layout)
      regsub -all -- {(^\n)|(\n$)|(\n +$)} $options(layout) ""   options(layout)
   }

   # Call appropriate procedure according to specified mode.
   switch -- $mode {
      add         {gpAdd}
      button      {gpButton}
      checkbutton {gpCheckbutton}
      clear       {gpClear}
      entry       {gpEntry}
      goto        {gpGoto}
      grid        {gpGrid}
      layout      {gpLayout}
      line        {gpLine}
      link        {gpLink}
      menu        {gpMenu}
      menubutton  {gpMenubutton}
      notebook    {gpNotebook}
      radiobutton {gpRadiobutton}
      set         {gpSet}
      style       {gpStyle}
      tablelist   {gpTablelist}
      text        {gpText}
      tree        {gpTree}
      window      {gpWindow}
   }

}

#=======================================================================#
# PROC   : ::gridplus::gpAdd                                            #
# PURPOSE: Add non-gridplus widget to group.                            #
#=======================================================================#

proc ::gridplus::gpAdd {} {
   upvar 1 options options

   variable gpGroup

   set gpGroup($options(name)) $options(-group)

}

#=======================================================================#
# PROC   : ::gridplus::gpButton                                         #
# PURPOSE: Create button grid.                                          #
#=======================================================================#

proc ::gridplus::gpButton {} {
   upvar 1 options options

   variable gpFocus
   variable gpGroup
   variable gpGroupState

   set foregroundColor [lindex [split $options(-buttoncolor) /] 0]
   set backgroundColor [lindex [split $options(-buttoncolor) /] 1]

   if {$foregroundColor eq ""} {set foregroundColor "black"}
   if {$backgroundColor eq ""} {set backgroundColor [. cget -background]}

   set gridData {}
   set rowCount 0

   if {! [regexp -- {^[.]([^.]+)[.]} $options(name) dummy window]} {
      set window {}
   }

   foreach row [split $options(layout) "\n"] {
      set columnCount 0
      foreach column $row {
         set buttonHelp   {}
         set buttonText   {}
         set createButton 0
         set gridColumn   {}
         set icon         0
         set size         $options(-size)
         set state        $options(-state)
         set validate     0
         foreach item $column {
            switch -regexp -- $item {
               ^[.] {
                  set createButton 1
                  if {! [regexp {(^[^:]+)(:[nsewc]+$)} $item dummy1 item sticky]} {set sticky {}}
                  set buttonName   $options(name),[string range $item 1 end]
                  if {$options(-group) ne ""} {set gpGroup($buttonName) $options(-group)}  
                  lappend gridColumn $buttonName$sticky
               }
               ^: {
                  set icon         1
                  set buttonIcon   [string range $item 1 end]
                  if {! $createButton} {
                     set createButton 1
                     set buttonName   $options(name),$buttonIcon
                     if {$options(-group) ne ""} {set gpGroup($buttonName) $options(-group)}  
                     lappend gridColumn $buttonName
                  }
               }
               ^[0-9]+$ {
                  set size $item
               }
               ^@ {
                  set gridName .[string range $item 1 end]
                  lappend gridColumn $gridName
               }
               ^% {
                  set gpGroup($buttonName) [string range $item 1 end]
               }
               ^[|]$ {
                  lappend gridColumn $item
               }
               ^[?] {
                  set buttonHelp [mc [string range $item 1 end]]
               }
               ^<$ {
                  set state disabled
               }
               ^>$ {
                  set state normal
               }
               ^!$ {
                  set validate 1
               }
               default {
                  set buttonText [mc $item]
               }
            }
         }

         if {$createButton} {
            if {[info exists gpGroup($buttonName)] && [info exists gpGroupState($gpGroup($buttonName))]} {
               set state $gpGroupState($gpGroup($buttonName))
            }

            if {$options(-proc)} {
               set command "set gridplus::gpFocus \[focus\];gpProc $buttonName"
            } else {
               set command "set gridplus::gpFocus \[focus\];$options(-prefix)$buttonName"
               regsub -all {[.]} $command ":" command
               regsub      {;:}  $command ";" command
            }

            if {$icon} {
               ::icons::icons create -file [file join $options(-iconpath) $options(-iconfile)] $buttonIcon
               if {$buttonText eq ""} {
                  button $buttonName -background $backgroundColor -command "::gridplus::gpCommand {$command} .$window $validate" -image ::icon::$buttonIcon -relief $options(-buttonrelief) -state $state -takefocus $options(-takefocus)
               } else {
                  button $buttonName -background $backgroundColor -command "::gridplus::gpCommand {$command} .$window $validate" -image ::icon::$buttonIcon -relief $options(-buttonrelief) -state $state -takefocus $options(-takefocus) -text $buttonText -width $size -compound $options(-compound)
               }
            } else {
               button $buttonName -background $backgroundColor -foreground $foregroundColor -command "::gridplus::gpCommand {$command} .$window $validate" -relief $options(-buttonrelief) -state $state -takefocus $options(-takefocus) -text $buttonText -width $size
            }
            if {$state eq "disabled"} {$buttonName configure -takefocus 0}
            if {$buttonHelp ne ""} {
               gpButtonHelpInit $buttonName $buttonHelp
            }
         }
         lappend gridData $gridColumn
         incr columnCount
      }
      lappend gridData !!!!
      incr rowCount
   }

   regsub -all {!!!!} $gridData \n gridData 

   set gridCommand "::gridplus::gridplus grid $options(name)"

   foreach option [lsort [array names options -*]] {
      set gridCommand "$gridCommand $option {$options($option)}"
   }

   set gridCommand "$gridCommand {$gridData}"

   eval $gridCommand
}

#=======================================================================#
# PROCS  : ::gridplus::gpButtonHelpInit                                 #
#        : ::gridplus::gpButtonHelpDelay                                #
#        : ::gridplus::gpButtonHelpCancel                               #
#        : ::gridplus::gpButtonHelpShow                                 #
# PURPOSE: Gridplus button help.                                        #
#=======================================================================#

proc ::gridplus::gpButtonHelpInit {item message} {
   variable gpButtonHelp 

   if {! [winfo exists .gpButtonHelp]} {
      toplevel .gpButtonHelp -background black -borderwidth 1 -relief flat
      label    .gpButtonHelp.message -background lightyellow
      pack     .gpButtonHelp.message
      wm overrideredirect .gpButtonHelp 1
      wm withdraw         .gpButtonHelp
   }

   set gpButtonHelp($item) $message
   bind $item <Enter> {::gridplus::gpButtonHelpDelay %W}
   bind $item <Leave> {::gridplus::gpButtonHelpCancel}
}

proc ::gridplus::gpButtonHelpDelay {item} {
   variable gpButtonHelp
 
   gpButtonHelpCancel
   set gpButtonHelp(delay) [after 300 [list ::gridplus::gpButtonHelpShow $item]]
}

proc ::gridplus::gpButtonHelpCancel {} {
   variable gpButtonHelp
 
   if {[info exists gpButtonHelp(delay)]} {
      after cancel $gpButtonHelp(delay)
      unset gpButtonHelp(delay)
   }

   if {[winfo exists .gpButtonHelp]} {
      wm withdraw .gpButtonHelp
   }
}

proc ::gridplus::gpButtonHelpShow {item} {
   variable gpButtonHelp
 
   .gpButtonHelp.message configure -text $gpButtonHelp($item)
 
   set helpX [expr [winfo rootx $item] + 10]
   set helpY [expr [winfo rooty $item] + [winfo height $item]]
 
   wm geometry  .gpButtonHelp +$helpX+$helpY
   wm deiconify .gpButtonHelp
 
   raise .gpButtonHelp
 
   unset gpButtonHelp(delay)
}

#=======================================================================#
# PROC   : ::gridplus::gpCheckbutton                                    #
# PURPOSE: Create checkbutton grid.                                     #
#=======================================================================#

proc ::gridplus::gpCheckbutton {} {
   upvar 1 options options

   global {}

   variable gpGroup
   variable gpGroupState

   set gridData {}
   set rowCount 0

   foreach row [split $options(layout) "\n"] {
      set columnCount 0
      foreach column $row {
         set createCheckbutton 0
         set gridColumn        {}
         set select            0
         set state             $options(-state)
         foreach item $column {
            switch -glob -- $item {
               .* {
                  set createCheckbutton 1
                  if {! [regexp {(^[^:]+)(:[nsewc]+$)} $item dummy1 item sticky]} {set sticky {}}
                  set checkbuttonName   $options(name),[string range $item 1 end]
                  if {$options(-group) ne ""} {set gpGroup($checkbuttonName) $options(-group)}  
                  lappend gridColumn $checkbuttonName$sticky
               }
               @* {
                  set gridName .[string range $item 1 end]
                  lappend gridColumn $gridName
               }
               %* {
                  set gpGroup($checkbuttonName) [string range $item 1 end]
               }
               +* {
                  set select 1
               }
               -* {
                  set select 0
               }
               default {
                  lappend gridColumn $item
               }
            }
         }

         if {$createCheckbutton} {
            if {[info exists gpGroup($checkbuttonName)] && [info exists gpGroupState($gpGroup($checkbuttonName))]} {
               set state $gpGroupState($gpGroup($checkbuttonName))
            }
            set ($checkbuttonName) {}
            checkbutton $checkbuttonName -background $options(-background) -offvalue 0 -onvalue 1 -state $state -takefocus $options(-takefocus) -variable ($checkbuttonName)
            if {$state eq "disabled"} {$checkbuttonName configure -takefocus 0}
            if {$select} {
               $checkbuttonName select
            } else {
               $checkbuttonName deselect
            }
         }

         lappend gridData $gridColumn
         incr columnCount
      }
      lappend gridData !!!!
      incr rowCount
   }

   regsub -all {!!!!} $gridData \n gridData 

   set gridCommand "::gridplus::gridplus grid $options(name)"

   foreach option [array names options -*] {
      set gridCommand "$gridCommand $option {$options($option)}"
   }

   set gridCommand "$gridCommand {$gridData}"

   eval $gridCommand
}

#=======================================================================#
# PROC   : ::gridplus::gpClear                                          #
# PURPOSE: Clear window and unset associated variables.                 #
#=======================================================================#

proc ::gridplus::gpClear {} {
   upvar 1 options options

   global {}

   variable gpButtonHelp
   variable gpGroup
   variable gpInfo
   variable gpTabOrder
   variable gpValidations

   $options(name) configure -menu {}

   unset -nocomplain gpInfo(validation:failed)
   unset -nocomplain gpValidations($options(name))

   foreach item [winfo child $options(name)] {
      set class [winfo class $item]

      if {[regexp -- {^[.]_} $item]} {
         continue
      }

      if {$class ne "Toplevel"} {
         if {$options(-variables) && [info exists ($item)]} {
            if {$class eq "Entry"} {
               $item configure -textvariable {}
            }
            unset ($item)
         }
         if {[info exists gpButtonHelp($item)]} {
            unset gpButtonHelp($item)
         }
         foreach infoItem [array names gpInfo $item:*] {
            unset gpInfo($infoItem)
         }
         foreach tabOrderItem [array names gpTabOrder $item:*] {
            unset gpTabOrder($tabOrderItem)
         }
         if {[info exists gpGroup($item)]} {
            unset gpGroup($item)
         }
         if {$class eq "Menu"} {
            foreach menuGroupItem [array names gpGroup $item.*] {
               unset gpGroup($menuGroupItem)
            }
         }
         destroy $item
      }
   }
}

#=======================================================================#
# PROC   : ::gridplus::gpCommand                                        #
# PURPOSE: Evals command, performing validations if required.           #
#=======================================================================#

proc ::gridplus::gpCommand {command window validate} {

   variable gpValidations

   if {$validate && [info exists gpValidations($window)]} {
      foreach validationInfo $gpValidations($window) {
         set entry [lindex [split $validationInfo :] 0]
         regexp -- {:(.+)$} $validationInfo dummy validation
         if {! [::gridplus::gpValidate $entry $validation]} {
            ::gridplus::gpValidateFailed $entry
            return
         }
      }
   }

   eval $command
}

#=======================================================================#
# PROC   : ::gridplus::gpCreateIcons                                    #
# PURPOSE: Creates default icons for GRIDPLUS Tree.                     #
#=======================================================================#

proc ::gridplus::gpCreateIcons {} {

   switch -- [gpGetOption Gridplus.treeActionStyle raised] {
      arrow {
         image create photo ::icon::contract -data {
            R0lGODlhCQAJAIAAAPz+/AQCBCH5BAEAAAAALAAAAAAJAAkAAAIMhI+pG+1x
            HkSOrXsLACH+aENyZWF0ZWQgYnkgQk1QVG9HSUYgUHJvIHZlcnNpb24gMi41
            DQqpIERldmVsQ29yIDE5OTcsMTk5OC4gQWxsIHJpZ2h0cyByZXNlcnZlZC4N
            Cmh0dHA6Ly93d3cuZGV2ZWxjb3IuY29tADs=
         }
         image create photo ::icon::expand -data {
            R0lGODlhCQAJAIAAAPz+/AQCBCH5BAEAAAAALAAAAAAJAAkAAAIPhI8WebHs
            HopSOVgb26YAACH+aENyZWF0ZWQgYnkgQk1QVG9HSUYgUHJvIHZlcnNpb24g
            Mi41DQqpIERldmVsQ29yIDE5OTcsMTk5OC4gQWxsIHJpZ2h0cyByZXNlcnZl
            ZC4NCmh0dHA6Ly93d3cuZGV2ZWxjb3IuY29tADs=
         }
      }
      flat {
         image create photo ::icon::contract -data {
            R0lGODlhCQAJAIEAAISChPz+/AQCBAAAACH5BAAAAAAALAAAAAAJAAkAAAIR
            hI+hG8brXgNBWDtllvGhbxQAIf5oQ3JlYXRlZCBieSBCTVBUb0dJRiBQcm8g
            dmVyc2lvbiAyLjUNCqkgRGV2ZWxDb3IgMTk5NywxOTk4LiBBbGwgcmlnaHRz
            IHJlc2VydmVkLg0KaHR0cDovL3d3dy5kZXZlbGNvci5jb20AOw==
         }
         image create photo ::icon::expand -data {
            R0lGODlhCQAJAIEAAISChPz+/AQCBAAAACH5BAAAAAAALAAAAAAJAAkAAAIU
            hI+hG7aynILKiXsbeHHPznTIWAAAIf5oQ3JlYXRlZCBieSBCTVBUb0dJRiBQ
            cm8gdmVyc2lvbiAyLjUNCqkgRGV2ZWxDb3IgMTk5NywxOTk4LiBBbGwgcmln
            aHRzIHJlc2VydmVkLg0KaHR0cDovL3d3dy5kZXZlbGNvci5jb20AOw==
         }
      }
      default {
         image create photo ::icon::contract -data {
            R0lGODlhCQAJAIIAAFxaXPz+/KSipMTCxAQCBAAAAAAAAAAAACH5BAAAAAAA
            LAAAAAAJAAkAAAMYCLoc/kKFQWkEs44biPectkniJZzoxawJACH+aENyZWF0
            ZWQgYnkgQk1QVG9HSUYgUHJvIHZlcnNpb24gMi41DQqpIERldmVsQ29yIDE5
            OTcsMTk5OC4gQWxsIHJpZ2h0cyByZXNlcnZlZC4NCmh0dHA6Ly93d3cuZGV2
            ZWxjb3IuY29tADs=
         }
         image create photo ::icon::expand -data {
            R0lGODlhCQAJAIIAAFxaXPz+/KSipMTCxAQCBAAAAAAAAAAAACH5BAAAAAAA
            LAAAAAAJAAkAAAMaCLoc/kKFMQiNYNYrifeYZg0hNWJCqmJMmwAAIf5oQ3Jl
            YXRlZCBieSBCTVBUb0dJRiBQcm8gdmVyc2lvbiAyLjUNCqkgRGV2ZWxDb3Ig
            MTk5NywxOTk4LiBBbGwgcmlnaHRzIHJlc2VydmVkLg0KaHR0cDovL3d3dy5k
            ZXZlbGNvci5jb20AOw==
         }
      }
   }

   image create photo ::icon::file -data {
      R0lGODlhEAAQAIIAAPwCBFxaXISChPz+/MTCxKSipAAAAAAAACH5BAEAAAAA
      LAAAAAAQABAAAANCCLrcGzBC4UAYOE8XiCdYF1BMJ5ye1HTfNxTBSpy0QMBy
      ++HlXNu8h24X6/2AReHwllRcMtCgs0CtVpsWiRZbqfgTACH+aENyZWF0ZWQg
      YnkgQk1QVG9HSUYgUHJvIHZlcnNpb24gMi41DQqpIERldmVsQ29yIDE5OTcs
      MTk5OC4gQWxsIHJpZ2h0cyByZXNlcnZlZC4NCmh0dHA6Ly93d3cuZGV2ZWxj
      b3IuY29tADs=
   }

   image create photo ::icon::folder -data {
      R0lGODlhEAAQAIIAAPwCBFxaXMTCxPz+/KSipAAAAAAAAAAAACH5BAEAAAAA
      LAAAAAAQABAAAAM3CLrc/i/IAFcQWFAos56TNYxkOWhKcHossals+64x5qZ0
      fQNwbc++Hy4o2F0IyKTSCGqCKhB/AgAh/mhDcmVhdGVkIGJ5IEJNUFRvR0lG
      IFBybyB2ZXJzaW9uIDIuNQ0KqSBEZXZlbENvciAxOTk3LDE5OTguIEFsbCBy
      aWdodHMgcmVzZXJ2ZWQuDQpodHRwOi8vd3d3LmRldmVsY29yLmNvbQA7
   }

   image create photo ::icon::spacer -data {
      R0lGODlhCQAJAIAAAPwCBAAAACH5BAEAAAAALAAAAAAJAAkAAAIIhI+py+0P
      UQEAIf5oQ3JlYXRlZCBieSBCTVBUb0dJRiBQcm8gdmVyc2lvbiAyLjUNCqkg
      RGV2ZWxDb3IgMTk5NywxOTk4LiBBbGwgcmlnaHRzIHJlc2VydmVkLg0KaHR0
      cDovL3d3dy5kZXZlbGNvci5jb20AOw==
   }
}

#=======================================================================#
# PROC   : ::gridplus::gpEntry                                          #
# PURPOSE: Create entry grid.                                           #
#=======================================================================#

proc ::gridplus::gpEntry {} {
   upvar 1 options options

   global {}

   variable gpGroup
   variable gpGroupState
   variable gpValidations

   set gridData {}
   set rowCount 0

   if {! [regexp -- {^[.]([^.]+)[.]} $options(name) dummy window]} {
      set window {}
   }

   foreach row [split $options(layout) "\n"] {
      set columnCount 0
      foreach column $row {
         set action      0
         set createEntry 0
         set focus       0
         set gridColumn  {}
         set secret      0
         set size        $options(-size)
         set state       $options(-state)
         set validate    none
         set validation  {}

         foreach item $column {
            switch -regexp -- $item {
               ^[.] {
                  set createEntry 1
                  if {! [regexp {(^[^:]+)(:[nsewc]+$)} $item dummy1 item sticky]} {set sticky {}}
                  set entryName   $options(name),[string range $item 1 end]
                  if {$options(-group) ne ""} {set gpGroup($entryName) $options(-group)}  
                  lappend gridColumn $entryName$sticky
               }
               ^[0-9]+$ {
                  set size $item
               }
               ^@ {
                  set gridName .[string range $item 1 end]
                  lappend gridColumn $gridName
               }
               ^% {
                  set gpGroup($entryName) [string range $item 1 end]
               }
               ^[+]$ {
                  set focus 1
               }
               ^! {
                  set validation [string range $item 1 end]
                  set validate   focusout
                  lappend gpValidations(.$window) $entryName:$validation
               }
               ^<$ {
                  set state disabled
               }
               ^>$ {
                  set state normal
               } 
               ^[*]$ {
                  set secret 1
               }
               ^~ {
                  set action 1
                  set command [string range $item 1 end]
               }
               default {
                  lappend gridColumn $item

               }
            }
         }

         if {$createEntry} {
            if {[info exists gpGroup($entryName)] && [info exists gpGroupState($gpGroup($entryName))]} {
               set state $gpGroupState($gpGroup($entryName))
            }
            set ($entryName) {}
            entry $entryName -disabledforeground black -invcmd "::gridplus::gpValidateFailed %W" -state $state -takefocus $options(-takefocus) -textvariable ($entryName) -validate $validate -vcmd "::gridplus::gpValidate %W $validation" -width $size
            if {$state eq "disabled"} {
               $entryName configure -background lightgray -takefocus 0
            }
            if {$action} {
               if {$command eq ""} {
                  set command $entryName
               }
               if {$options(-proc)} {
                  set command "gpProc $command"
               } else {
                  set command "$options(-prefix)$command"
                  regsub -all {[.]} $command ":" command
                  regsub      {^:}  $command {}  command 
               }
               bind $entryName <Return> $command
            }
            if {$secret} {$entryName configure -show "*"}
            if {$focus}  {focus $entryName}
         }

         lappend gridData $gridColumn
         incr columnCount
      }
      lappend gridData !!!!
      incr rowCount
   }

   regsub -all {!!!!} $gridData \n gridData 

   set gridCommand "::gridplus::gridplus grid $options(name)"

   foreach option [array names options -*] {
      set gridCommand "$gridCommand $option {$options($option)}"
   }

   set gridCommand "$gridCommand {$gridData}"

   eval $gridCommand
}

#=======================================================================#
# PROC   : ::gridplus::gpGetOption                                      #
# PURPOSE: Get option from "option database". Use default if not found. #
#=======================================================================#

proc ::gridplus::gpGetOption {option default} {

   set value [option get . $option -]

   if {$value eq ""} {
      return $default
   } else {
      return $value
   }
}

#=======================================================================#
# PROC   : ::gridplus::gpGoto                                           #
# PURPOSE: Move text widget display to specified label.                 #
#=======================================================================#

proc ::gridplus::gpGoto {} {
   upvar 1 options options

   global {}

   $options(name).text yview $options(layout)

   set ($options(name)) $options(layout)
}

#=======================================================================#
# PROC   : ::gridplus::gpGrid                                           #
# PURPOSE: Create grid.                                                 #
#=======================================================================#

proc ::gridplus::gpGrid {} {
   upvar 1 options options

   global {}

   variable gpTabOrder

   set labelColor(1) [lindex [split $options(-labelcolor) /] 0]
   set labelColor(2) [lindex [split $options(-labelcolor) /] 1]
   set labelStyle(1) [lindex [split $options(-labelstyle) /] 0]
   set labelStyle(2) [lindex [split $options(-labelstyle) /] 1]

   regsub -all -- {,} $labelStyle(1) { } labelStyle(1)
   regsub -all -- {,} $labelStyle(2) { } labelStyle(2)

   if {! [string match */* $options(-labelcolor)]} {set labelColor(2) $labelColor(1)}

   if {$labelColor(1) eq ""} {set labelColor(1) "black"}
   if {$labelColor(2) eq ""} {set labelColor(2) "black"}

   labelframe $options(name) -background $options(-background) -borderwidth $options(-borderwidth) -padx $options(-padx) -pady $options(-pady) -relief $options(-relief) -text [mc $options(-title)]
   
   set rowID    0
   set rowTotal [llength [split $options(layout) "\n"]]
   set rowCount 1

   if {! [regexp -- {^[.]([^.]+)[.]} $options(name) dummy window]} {
      set window {}
   }

   if {$options(-subst)} {
      set options(layout) [subst -nobackslashes -nocommands $options(layout)]
   }

   foreach row [split $options(layout) "\n"] {
      set columnID    0
      set columnTotal [llength $row]
      set columnCount 1

      foreach column $row {
         switch -- [llength $column] {
            0 {
               set columnSpan 2
               set column "{}"
            }
            1 {
               set columnSpan 2
            }
            2 {
               set columnSpan 1
            }
            default {
               error "GRIDPLUS ERROR: Too many items in column"
            }
         }

         set columnItem 1

         foreach item $column {
            set bold      0
            set command   {}
            set labelIcon {}
            set sticky    {}
            set validate  0

            if {! [string match "*: " $item]} {
               regexp {(^[^:]+)(:([nsewc]+$)?)} $item dummy1 item dummy2 sticky
            }

            switch -- $sticky {
               c  {set sticky {}}
               "" {set sticky w}
            }

            switch -glob -- $item {
               .* {
                  set itemName $item
                  if {! [winfo exists $item]} {
                     set   itemName $options(name),[string range $item 1 end]
                     label $itemName -background $options(-background) -foreground $labelColor($columnItem) -justify $options(-justify) -wraplength $options(-wraplength) -textvariable ($itemName)
                     $itemName configure -font [gpSetFont [$itemName cget -font] $labelStyle($columnItem)]
                  }
                  grid configure $itemName -in $options(name) -column $columnID -row $rowID -columnspan $columnSpan -sticky $sticky
                  if {$options(-taborder) eq "column"} {
                     set gpTabOrder([format "%s:%03d%03d%03d" $options(name) $columnCount $rowCount $columnItem]) $itemName
                  } else {
                     set gpTabOrder([format "%s:%03d%03d%03d" $options(name) $rowCount $columnCount $columnItem]) $itemName
                  }
               }
               | {
                  frame $options(name).frame:$rowID:$columnID -background $options(-background) -borderwidth 2 -height 2 -relief sunken -width 2
                  grid configure $options(name).frame:$rowID:$columnID -in $options(name) -column $columnID -row $rowID -columnspan $columnSpan -sticky ns
               }
               :* {
                  if {! [regexp -- {^:([^:]*):([^:]*):([^:]*)$} $item dummy labelIcon command validate]} {
                     set labelIcon [string range $item 1 end]
                     regsub -- {%%$} $labelIcon {} labelIcon
                  }
                  if {$labelIcon eq ""} {
                     set labelIcon $options(-icon)
                  }
                  ::icons::icons create -file [file join $options(-iconpath) $options(-iconfile)] $labelIcon
                  label $options(name).label:$rowID:$columnID -background $options(-background) -image ::icon::$labelIcon
                  grid configure $options(name).label:$rowID:$columnID -in $options(name) -column $columnID -row $rowID -columnspan $columnSpan -sticky $sticky
                  if {$command ne ""} {
                     if {$options(-proc)} {
                        set command "set gridplus::gpFocus \[focus\];gpProc $command"
                     } else {
                        set command "set gridplus::gpFocus \[focus\];$options(-prefix)$command"
                        regsub -all {[.]} $command ":" command
                        regsub      {;:}  $command ";" command
                     }

                     bind $options(name).label:$rowID:$columnID <ButtonRelease-1> "eval \"::gridplus::gpCommand {$command} .$window $validate\""
                  }
               }
               default {
                  if {[string match ^* $item]} {
                     set bold 1
                     set item [string range $item 1 end]
                  }

                  regsub -all -- " +\n +" $item "\n" item
                  label $options(name).label:$rowID:$columnID -background $options(-background) -foreground $labelColor($columnItem) -justify $options(-justify) -wraplength $options(-wraplength) -text [mc $item]
                  $options(name).label:$rowID:$columnID configure -font [gpSetFont [$options(name).label:$rowID:$columnID cget -font] $labelStyle($columnItem)]

                  if {$bold} {
                     set font "[$options(name).label:$rowID:$columnID cget -font] bold"
                     $options(name).label:$rowID:$columnID configure -font $font
                  }

                  grid configure $options(name).label:$rowID:$columnID -in $options(name) -column $columnID -row $rowID -columnspan $columnSpan -sticky $sticky
               }
            }
            incr columnID $columnSpan
            incr columnItem
         }

         if {$columnCount != $columnTotal} {
            frame $options(name).space:$rowID:$columnID -background $options(-background) -width $options(-space)
            grid $options(name).space:$rowID:$columnID -column $columnID -row $rowID -sticky ew
            grid columnconfigure $options(name) $columnID -weight 1
            incr columnID
         }
         incr columnCount
      }

      incr rowID

      if {$rowCount != $rowTotal} {
         frame $options(name).space:$rowID:$columnID -background $options(-background) -height 4 -width 4
         grid $options(name).space:$rowID:$columnID -row $rowID -column 0 -sticky ns -columnspan 3
         grid rowconfigure $options(name) $rowID -weight 1
         incr rowID
      }

      incr rowCount
   }

   gpSetTabOrder $options(name)
}

#=======================================================================#
# PROC   : ::gridplus::gpInit                                           #
# PURPOSE: Gridplus initailise.                                         #
#=======================================================================#

proc ::gridplus::gpInit {} {
   variable gpConfig
   variable gpDateFormat
   variable gpValidation
   variable gpValidationText

   wm resizable . 0 0

   if {[namespace exists "::starkit"]} {
      set iconPath [file join $::starkit::topdir lib]
   } else {
      set iconPath [file join [info library]]
   }

   array set gpConfig [list                                  \
      dateformat   [gpGetOption Gridplus.dateFormat us]      \
      errormessage [gpGetOption Gridplus.errorMessage %]     \
      iconfile     [gpGetOption Gridplus.iconFile tkIcons]   \
      iconpath     [gpGetOption Gridplus.iconPath $iconPath] \
      prefix       [gpGetOption Gridplus.prefix {}]          \
      proc         [gpGetOption Gridplus.proc 0]             \
   ]

   switch -- $gpConfig(dateformat) {
      eu {
         set gpDateFormat(day)       0
         set gpDateFormat(month)     1
         set gpDateFormat(separator) .
      }
      uk {
         set gpDateFormat(day)       0
         set gpDateFormat(month)     1
         set gpDateFormat(separator) /
      }
      us {
         set gpDateFormat(day)       1
         set gpDateFormat(month)     0
         set gpDateFormat(separator) /
      }
   }

   set gpDateFormat(century) [gpGetOption Gridplus.century 50]

   array set gpValidation {
      alpha     {^[a-zA-Z]+$}
      alphanum  {^[a-zA-Z0-9]+$}
      date      {proc:gpValidateDate}
      money     {^[0-9]+[.][0-9][0-9]$}
      num       {^[0-9]+[.]?[0-9]*$}
      notnull   {[^\000]}
      int       {^[0-9]+$}
   }

   array set gpValidationText {
      alpha     {Alpha}
      alphanum  {Alphanumeric}
      date      {Date}
      money     {Money Format}
      notnull   {Not Null}
      num       {Numeric}
      int       {Integer}
   }

   ::gridplus::gpCreateIcons
}

#=======================================================================#
# PROC   : ::gridplus::gpInsertText                                     #
# PURPOSE: Inserts "tagged" data into text widget.                      #
#=======================================================================#

proc ::gridplus::gpInsertText {name tag end parameter position text} {
   upvar 1 options options

   global {}

   variable gpInfo

   if {! [regexp -- {^[.]([^.]+)[.]} $name dummy window]} {
      set window {}
   }

   set command        false
   set imageCommand   {}
   set imageInfo      {}
   set imageLink      {}
   set imageParameter {}
   set link           false
   set bgColor        $gpInfo($name:bgcolor)
   set fgColor        $gpInfo($name:fgcolor)
   set linkColor      $gpInfo($name:link)
   set setCommand     0
   set validate       0

   switch -- $end$tag {
      init     {set gpInfo($name:font)       $gpInfo($name:defaultfont)
                set gpInfo($name:size)       10
                set gpInfo($name:weight)     normal
                set gpInfo($name:slant)      roman
                set gpInfo($name:underline)  false}
      b        {set gpInfo($name:weight)     bold}
      /b       {set gpInfo($name:weight)     normal}
      bgcolor  {set bgColor                  [lindex [split $parameter :] 0]
                set bgParameter              [lindex [split $parameter :] 1]
                if {$bgParameter eq "default"} {set gpInfo($name:defaultbg) $bgColor}
                set gpInfo($name:bgcolor)    $bgColor}
      /bgcolor {set bgColor                  $gpInfo($name:defaultbg)
                set gpInfo($name:bgcolor)    $gpInfo($name:defaultbg)}
      color    {set fgColor                  [lindex [split $parameter :] 0]
                set fgParameter              [lindex [split $parameter :] 1]
                if {$fgParameter eq "default"} {set gpInfo($name:defaultfg) $fgColor}
                set gpInfo($name:fgcolor)    $fgColor}
      /color   {set fgColor                  $gpInfo($name:defaultfg)
                set gpInfo($name:fgcolor)    $gpInfo($name:defaultfg)}
      command  {set fgColor                  $gpInfo($name:normalcolor)
                set gpInfo($name:underline)  $gpInfo($name:normalstyle)
                set command                  [lindex [split $parameter :] 0]
                set commandParameter         [lindex [split $parameter :] 1]
                if {$commandParameter eq ""} {set commandParameter $text}}
      font     {set font                     [lindex [split $parameter :] 0]
                set fontParameter            [lindex [split $parameter :] 1]
                if {$fontParameter eq "default"} {set gpInfo($name:defaultfont) $font}
                set gpInfo($name:font)       $font}
      /font    {set gpInfo($name:font)       $gpInfo($name:defaultfont)}
      i        {set gpInfo($name:slant)      italic}
      /i       {set gpInfo($name:slant)      roman}
      image    {set imageInfo                $parameter}
      indent   {set gpInfo($name:indent)     $parameter
                set tabs [string repeat "\t" $parameter]
                set text "$tabs$text"}
      /indent  {set gpInfo($name:indent)     0}
      label    {set label                    [lindex [split $parameter :] 0]
                set labelParameter           [lindex [split $parameter :] 1]
                if {$labelParameter eq "default"} {set ($name) $label} 
                $name.text mark set $label "insert wordstart"
                $name.text mark gravity $label left}
      link     {set fgColor                  $gpInfo($name:normalcolor)
                set gpInfo($name:underline)  $gpInfo($name:normalstyle)
                set link                     $parameter}
      size     {set size                     [lindex [split $parameter :] 0]
                set sizeParameter            [lindex [split $parameter :] 1]
                if {$sizeParameter eq "default"} {set gpInfo($name:defaultsize) $size}
                set gpInfo($name:size)       [gridplus::gpSetFontSize $gpInfo($name:defaultsize) $size]}
      /size    {set gpInfo($name:size)       $gpInfo($name:defaultsize)}
      tab      {if {$parameter eq ""} {set parameter 1}
                set tabs [string repeat "\t" $parameter]
                set text "$tabs$text"}
      u        {set gpInfo($name:underline)  true}
      /u       {set gpInfo($name:underline)  false}
   }

   set tagName "tag[incr gpInfo($name:tagid)]"
   set font    "-family $gpInfo($name:font) -size $gpInfo($name:size) -slant $gpInfo($name:slant) -underline $gpInfo($name:underline) -weight $gpInfo($name:weight)"
   set indent  "[expr {$gpInfo($name:indent) * 0.5}]c"

   $name.text tag configure $tagName -lmargin1 $indent -lmargin2 $indent -background $bgColor -foreground $fgColor -font "$font"

   if {$imageInfo ne ""} {
      if {[string match *@* $imageInfo]} {
         set image          [lindex [split $imageInfo @] 0]
         set imageLink      [lindex [split $imageInfo @] 1]
      } else {
         set image          [lindex [split $imageInfo ~] 0]
         set imageCommand   [lindex [split [lindex [split $imageInfo ~] 1] :] 0]
         set imageParameter [lindex [split [lindex [split $imageInfo ~] 1] :] 1]

         if {$imageCommand ne ""} {
            set setCommand   1
            set imageCommand "$name,$imageCommand"

            if {$gpInfo($name:proc)} {
               set imageCommand "set gridplus::gpFocus \[focus\];gpProc $imageCommand"
            } else {
               set imageCommand "set gridplus::gpFocus \[focus\];$gpInfo($name:prefix)$imageCommand"
               regsub -all {[.]} $imageCommand ":" imageCommand
               regsub      {;:}  $imageCommand ";" imageCommand
            }
         }
      }

      if {[string match :* $image]} {
         set icon  [string range $image 1 end]
         set image "::icon::$icon"
         ::icons::icons create -file $gpInfo($name:iconlibrary) $icon
      }

      set imageName [$name.text image create end -image $image]

      $name.text tag add $imageName $imageName
      $name.text tag configure $imageName -background $bgColor

      if {$imageLink ne ""} {
         $name.text tag bind $imageName <Enter> "$name.text configure -cursor $gpInfo($name:linkcursor)"
         $name.text tag bind $imageName <Leave> "$name.text configure -cursor {}"
         $name.text tag bind $imageName <ButtonPress-1> "set ($name) $imageLink; $name.text yview $imageLink"
      } elseif {$setCommand} {
         $name.text tag bind $imageName <Enter> "$name.text configure -cursor $gpInfo($name:linkcursor)"
         $name.text tag bind $imageName <Leave> "$name.text configure -cursor {}"
         $name.text tag bind $imageName <ButtonPress-1> "set ($name) \"$imageParameter\"; ::gridplus::gpCommand {$imageCommand} .$window $validate"
      }
   }

   if {$command ne "false"} {

      set command "$name,$command"

      if {$gpInfo($name:proc)} {
         set command "set gridplus::gpFocus \[focus\];gpProc $command"
      } else {
         set command "set gridplus::gpFocus \[focus\];$gpInfo($name:prefix)$command"
         regsub -all {[.]} $command ":" command
         regsub      {;:}  $command ";" command
      }

      $name.text tag bind $tagName <Enter> "$name.text configure -cursor $gpInfo($name:linkcursor); $name.text tag configure $tagName -foreground $gpInfo($name:overcolor) -underline $gpInfo($name:overstyle)"
      $name.text tag bind $tagName <Leave> "$name.text configure -cursor {}; $name.text tag configure $tagName -foreground $gpInfo($name:normalcolor) -underline $gpInfo($name:normalstyle)"
      $name.text tag bind $tagName <ButtonPress-1> "set ($name) \"$commandParameter\"; ::gridplus::gpCommand {$command} .$window $validate"

      set gpInfo($name:underline) false
   }
   if {$link    ne "false"} {
      $name.text tag bind $tagName <Enter> "$name.text configure -cursor $gpInfo($name:linkcursor); $name.text tag configure $tagName -foreground $gpInfo($name:overcolor) -underline $gpInfo($name:overstyle)"
      $name.text tag bind $tagName <Leave> "$name.text configure -cursor {}; $name.text tag configure $tagName -foreground $gpInfo($name:normalcolor) -underline $gpInfo($name:normalstyle)"
      $name.text tag bind $tagName <ButtonPress-1> "set ($name) $link; $name.text yview $link"
      set gpInfo($name:underline) false
   }

   if {$text ne ""} {
      regsub -all {!b:}  $text "\u2022" text
      regsub -all {!ob:} $text \{       text
      regsub -all {!cb:} $text \}       text
      regsub -all {!bs:} $text {\\}     text
      regsub -all {!lt:} $text {<}      text
      regsub -all {!gt:} $text {>}      text
      $name.text insert $position $text $tagName 
   }
}

#=======================================================================#
# PROC   : ::gridplus::gpLayout                                         #
# PURPOSE: Create layout.                                               #
#=======================================================================#

proc ::gridplus::gpLayout {} {
   upvar 1 options options

   global {}

   variable gpTabOrder

   set rowCount      0
   set layout(items) {}
   set toplevel      {}

   if {$options(-subst)} {
      set options(layout) [subst -nobackslashes -nocommands $options(layout)]
   }

   foreach row [split $options(layout) "\n"] {
      set columnCount 0
      foreach column $row {
         set sticky {}
         regexp {(^[^:]+)(:([nsewc]+$)?)} $column dummy1 column dummy2 sticky
         switch -- $sticky {
            c  {set sticky {}}
            "" {set sticky w}
         }
         switch -glob -- $column {
            .* {
               set column [regsub -all -- {%} $column [string range $options(name) 1 end]]
               lappend layout(items) $column
               set layout(cell:$columnCount,$rowCount) $column
               set layout($column:x)      $columnCount
               set layout($column:y)      $rowCount
               set layout($column:xspan)  1
               set layout($column:yspan)  1
               set layout($column:sticky) $sticky
               if {$options(-taborder) eq "column"} {
                  set gpTabOrder([format "%s:%03d%03d001" $options(name) $columnCount $rowCount]) $column
               } else {
                  set gpTabOrder([format "%s:%03d%03d001" $options(name) $rowCount $columnCount]) $column
               }
            }
            - {
               if {$columnCount == 0} {error "GRIDPLUS ERROR (layout): Column span not valid in first column"}
               set previousColumn [expr {$columnCount - 1}]
               set cell $layout(cell:$previousColumn,$rowCount)
               set layout(cell:$columnCount,$rowCount) $layout(cell:$previousColumn,$rowCount)
               incr layout($cell:xspan)
            }
            ^ {
               if {$rowCount == 0} {error "GRIDPLUS ERROR (layout): Row span not valid in first row"}
               set previousRow [expr {$rowCount - 1}]
               set cell $layout(cell:$columnCount,$previousRow)
               set layout(cell:$columnCount,$rowCount) $layout(cell:$columnCount,$previousRow)
               incr layout($cell:yspan)
            }
            x {
            }
            default {
               error "GRIDPLUS ERROR (layout): Invalid item/option ($column)"
            }
         }
         incr columnCount
      }
      incr rowCount
   }

   if {[regexp {([^.]*)[.]main$} $options(name) match window]} {
      wm title .$window [mc $options(-title)]
      set options(-relief) flat
      set options(-title)  {}
   }

   labelframe $options(name) -background $options(-background) -borderwidth $options(-borderwidth) -relief $options(-relief) -text [mc $options(-title)]

   foreach item $layout(items) {
      set padxLeft  $options(-padx)
      set padxRight $options(-padx)

      if {$layout($item:x) == 0} {
         set padxLeft 0
      }
      if {[expr {$layout($item:x) + $layout($item:xspan)}] == $columnCount} {
         set padxRight 0
      }

      set padyTop    $options(-pady)
      set padyBottom $options(-pady)

      if {$layout($item:y) == 0} {
         set padyTop 0
      }
      if {[expr {$layout($item:y) + $layout($item:yspan)}] == $rowCount} {
         set padyBottom 0
      }

      set padx [list $padxLeft $padxRight]
      set pady [list $padyTop  $padyBottom]

      grid configure $item -in $options(name) -column $layout($item:x) -row $layout($item:y) -columnspan $layout($item:xspan) -rowspan $layout($item:yspan) -sticky $layout($item:sticky) -padx $padx -pady $pady
      grid columnconfigure $options(name) $layout($item:x) -weight 1
      grid rowconfigure    $options(name) $layout($item:y) -weight 1
      gpSetTabOrder $options(name)
   } 
}

#=======================================================================#
# PROC   : ::gridplus::gpLine                                           #
# PURPOSE: Gridplus create line.                                        #
#=======================================================================#

proc ::gridplus::gpLine {} {
   upvar 1 options options

   if {$options(-title) ne ""} {
      frame $options(name) -background $options(-background) -padx $options(-padx) -pady $options(-pady)
      frame $options(name).left  -background $options(-background) -borderwidth 2 -height 2 -relief sunken -width 5
      frame $options(name).right -background $options(-background) -borderwidth 2 -height 2 -relief sunken
      label $options(name).label -background $options(-background) -text [mc $options(-title)] -borderwidth 1
      grid configure $options(name).left  -column 0 -row 0 -sticky ew
      grid configure $options(name).label -column 1 -row 0
      grid configure $options(name).right -column 2 -row 0 -sticky ew
      grid columnconfigure $options(name) 2 -weight 1
   } else {
     frame $options(name) -background $options(-background) -borderwidth $options(-borderwidth) -height $options(-linewidth) -padx $options(-padx) -pady $options(-pady) -relief $options(-linerelief) -width $options(-width)
   }
}

#=======================================================================#
# PROC   : ::gridplus::gpLink                                           #
# PURPOSE: Create link grid.                                            #
#=======================================================================#

proc ::gridplus::gpLink {} {
   upvar 1 options options

   set normalColor [lindex [split $options(-linkcolor) /] 0]
   set overColor   [lindex [split $options(-linkcolor) /] 1]
   set normalStyle [lindex [split $options(-linkstyle) /] 0]
   set overStyle   [lindex [split $options(-linkstyle) /] 1]

   regsub -- {[&]} $overStyle $normalStyle, overStyle

   regsub -all -- {,} $normalStyle { } normalStyle
   regsub -all -- {,} $overStyle   { } overStyle

   if {! [string match */* $options(-linkcolor)]} {set overColor $normalColor}

   if {$normalColor eq ""} {set normalColor "black"}
   if {$overColor   eq ""} {set overColor "black"}

   set gridData {}
   set rowCount 0

   if {! [regexp -- {^[.]([^.]+)[.]} $options(name) dummy window]} {
      set window {}
   }

   foreach row [split $options(layout) "\n"] {
      set columnCount 0
      foreach column $row {
         set createLink   0
         set gridColumn   {}
         set icon         0
         set linkText     {}
         set validate     0
         foreach item $column {
            switch -regexp -- $item {
               ^[.] {
                  set createLink 1
                  if {! [regexp {(^[^:]+)(:[nsewc]+$)} $item dummy1 item sticky]} {set sticky {}}
                  set linkName   $options(name),[string range $item 1 end]
                  lappend gridColumn $linkName$sticky
               }
               ^[|]$ {
                  lappend gridColumn $item
               }
               ^[:] {
                  lappend gridColumn $item%%
                  set icon 1
               }
               ^[+]$ {
                  lappend gridColumn "\u2022"
               }
               ^[-]$ {
                  lappend gridColumn {}
               }
               ^[!]$ {
                  set validate 1
               }
               default {
                  set linkText [mc $item]
               }
            }
         }

         if {$createLink} {
            if {$icon} {
               regsub -- {(:[^:]*)%%} $gridColumn "\\1:$linkName:$validate" gridColumn
            }

            if {$options(-proc)} {
               set command "set gridplus::gpFocus \[focus\];gpProc $linkName"
            } else {
               set command "set gridplus::gpFocus \[focus\];$options(-prefix)$linkName"
               regsub -all {[.]} $command ":" command
               regsub      {;:}  $command ";" command
            }

            label $linkName -background $options(-background) -foreground $options(-foreground) -text [mc $linkText]

            set baseFont [$linkName cget -font]

            if {[string is integer -strict [lindex $normalStyle 0]]} {
               set normalFont "[list [lindex $baseFont 0]] $normalStyle"
            } else {
               set normalFont "$baseFont $normalStyle"
            }
            if {[string is integer -strict [lindex $overStyle 0]]} {
               set overFont "[list [lindex $baseFont 0]] $overStyle"
            } else {
               set overFont "$baseFont $overStyle"
            }

            $linkName configure -font $normalFont -foreground $normalColor

            bind $linkName <Enter>           "$linkName configure -font \"$overFont\" -foreground $overColor -cursor $options(-linkcursor)"
            bind $linkName <Leave>           "$linkName configure -font \"$normalFont\" -foreground $normalColor -cursor {}"
            bind $linkName <ButtonRelease-1> "eval \"::gridplus::gpCommand {$command} .$window $validate\""
         } else {
           lappend gridColumn $linkText
         }
         lappend gridData $gridColumn
         incr columnCount
      }
      lappend gridData !!!!
      incr rowCount
   }

   regsub -all {!!!!} $gridData \n gridData 

   set gridCommand "::gridplus::gridplus grid $options(name)"

   foreach option [lsort [array names options -*]] {
      set gridCommand "$gridCommand $option {$options($option)}"
   }

   set gridCommand "$gridCommand {$gridData}"

   eval $gridCommand
}

#=======================================================================#
# PROC   : ::gridplus::gpMenu                                           #
# PURPOSE: Create menu(bar).                                            #
#=======================================================================#

proc ::gridplus::gpMenu {} {
   upvar 1 options options

   if {$options(name) eq "."} {
      set rootMenu .menubar
      $options(name) configure -menu $rootMenu
   } elseif {[winfo exists $options(name)] && [winfo class $options(name)] eq "Toplevel"} {
      set rootMenu $options(name).menubar
      $options(name) configure -menu $rootMenu
   } else {
      set rootMenu $options(name)
   }

   menu $rootMenu

   $rootMenu configure -tearoff 0

   set rootMenuIndex 0

   foreach {menuLabel menuEntries} $options(layout) {
      if {$menuLabel eq "~"} {
         ::gridplus::gpMenuOption $rootMenu {} $rootMenuIndex $menuEntries
         incr rootMenuIndex
         continue  
      }

      if {[string match @* $menuEntries]} {
         set cascade ".[string range $menuEntries 1 end]"
         $rootMenu add cascade -label [mc $menuLabel] -menu $cascade
         continue
      }

      set menu [string tolower $menuLabel]

      $rootMenu add cascade -label [mc $menuLabel] -menu $rootMenu.$menu 
      menu $rootMenu.$menu
      $rootMenu.$menu configure -tearoff 0

      set menuIndex 0

      foreach menuEntryData $menuEntries {
         ::gridplus::gpMenuOption $rootMenu $menu $menuIndex $menuEntryData  
         incr menuIndex
      }

      incr rootMenuIndex
   }

}

#=======================================================================#
# PROC   : ::gridplus::gpMenuOption                                     #
# PURPOSE: Create menu(bar) option.                                     #
#=======================================================================#

proc ::gridplus::gpMenuOption {rootMenu menu menuIndex menuEntryData} {
   upvar 1 options options

   variable gpGroup
   variable gpGroupState

   set menuEntryLabel   [lindex $menuEntryData 0]
   set menuEntryOptions [lrange $menuEntryData 1 end]
   set menuEntry        [string tolower $menuEntryLabel]

   regsub -all -- { } $menuEntry {_} menuEntry

   if {$menuEntry eq "-"} {
      $rootMenu.$menu add separator     
   } else {
      if {$menu eq ""} {
         set command     $rootMenu,$menuEntry
         set menuEntryID $rootMenu@$menuIndex
         set menuName    {}
      } else {
         set command     $rootMenu:$menu,$menuEntry
         set menuEntryID $rootMenu.$menu@$menuIndex
         set menuName    .$menu
      }
      set cascade     {}
      set compound    none
      set menuIcon    {}
      set state       $options(-state)
      set validate    0

      foreach item $menuEntryOptions {
         switch -regexp -- $item {
            ^% {
               set gpGroup($menuEntryID) [string range $item 1 end]
            }
            ^<$ {
               set state disabled
            }
            ^>$ {
               set state normal
            }
            ^!$ {
               set validate 1
            }
            ^@ {
               set cascade ".[string range $item 1 end]"
            }
            ^[.] {
               set command [string range $item 1 end]
            }
            ^: {
               set menuIcon "::icon::[::icons::icons create -file [file join $options(-iconpath) $options(-iconfile)] [string range $item 1 end]]"
               set compound left
            }
         }
      }

      if {$options(-proc)} {
         set command "gpProc $command"
      } else {
         set command "$options(-prefix)$command"
         regsub -all {[.]} $command ":" command
         regsub      {^:}  $command {}  command
      }

      if {[info exists gpGroup($menuEntryID)] && [info exists gpGroupState($gpGroup($menuEntryID))]} {
         set state $gpGroupState($gpGroup($menuEntryID))
      }

      if {$cascade ne ""} {
         $rootMenu$menuName add cascade -label [mc $menuEntryLabel] -menu $cascade -state $state -compound $compound -image $menuIcon
      } else {
         $rootMenu$menuName add command  -label [mc $menuEntryLabel] -command "::gridplus::gpCommand {$command} $options(name) $validate" -state $state -compound $compound -image $menuIcon
      }
   }
}

#=======================================================================#
# PROC   : ::gridplus::gpNotebook                                       #
#        : ::gridplus::gpNotebookSet                                    #
# PURPOSE: Create notebook.                                             #
#=======================================================================#

proc ::gridplus::gpNotebook {} {
   upvar 1 options options

   set foregroundColor [lindex [split $options(-buttoncolor) /] 0]
   set backgroundColor [lindex [split $options(-buttoncolor) /] 1]

   if {$foregroundColor eq ""} {set foregroundColor "black"}
   if {$backgroundColor eq ""} {set backgroundColor [. cget -background]}

   set ID 0

   frame $options(name) -background $options(-background) -borderwidth $options(-borderwidth) -relief $options(-relief)
   frame $options(name).buttons -background $options(-background)
   frame $options(name).contents -background $options(-background) -relief $options(-irelief) -borderwidth 2
   frame $options(name).spacer -background $options(-background)

   foreach {button item} $options(layout) {
      incr ID
      button $options(name).buttons.button$ID -background $backgroundColor -foreground $foregroundColor -relief groove -takefocus 0 -text [mc $button] -command "::gridplus::gpNotebookSet $options(name) $ID $item $backgroundColor [list $options(layout)]"
      pack $options(name).buttons.button$ID -side left
   }

   grid $options(name).buttons  -row 0 -column 0 -sticky nw
   grid $options(name).contents -row 1 -column 0 -sticky nsew

   grid columnconfigure $options(name) 0 -weight 1
   grid rowconfigure    $options(name) 1 -weight 1

   after idle $options(name).buttons.button1 invoke
}

proc ::gridplus::gpNotebookSet {name setID setItem background layout} {
   global {}

   variable gpTabOrder

   set ID 0

   foreach {button item} $layout {
      incr ID
      if {$item eq $setItem} {
         grid configure $setItem     -in $name.contents -column 0 -row 0 -sticky nsw
         grid configure $name.spacer -in $name.contents -column 1 -row 0 -sticky nsew
         grid rowconfigure    $name.contents 0 -weight 1
         grid columnconfigure $name.contents 1 -weight 1
         set ($name) [$name.buttons.button$setID cget -text]
         $name.buttons.button$setID configure -background $background -relief raised
      } else {
         $name.buttons.button$ID configure -background gray60 -relief groove
         grid forget $item
      }
   }

   set gpTabOrder($name:000000) $setItem

   if {! [regexp {(^[.][^.]+)[.]} $setItem match window]} {
      set window {}
   }

   gpSetTabOrder $window.main
}

#=======================================================================#
# PROC   : ::gridplus::gpParseTags                                      #
# PURPOSE: Parse tags for text widget.                                  #
#=======================================================================#

proc ::gridplus::gpParseTags {name tagText position} {

   regsub -all \{   $tagText {!ob:} tagText
   regsub -all \}   $tagText {!cb:} tagText
   regsub -all {\\} $tagText {!bs:} tagText

   set whitespace " \t\r\n"
   set pattern <(/?)(\[^$whitespace>]+)\[$whitespace]*(\[^>]*)>

   set substitute "\}\n::gridplus::gpInsertText $name {\\2} {\\1} {\\3} $position \{"
   regsub -all $pattern $tagText $substitute tagText

   eval "::gridplus::gpInsertText $name {init} {} {} $position {$tagText}"
}

#=======================================================================#
# PROC   : ::gridplus::gpRadiobutton                                    #
# PURPOSE: Create radiobutton grid.                                     #
#=======================================================================#

proc ::gridplus::gpRadiobutton {} {
   upvar 1 options options

   global {}

   variable gpGroup
   variable gpGroupState

   set gridData {}
   set rowCount 0

   foreach row [split $options(layout) "\n"] {
      set columnCount 0
      foreach column $row {
         set createRadiobutton 0
         set gridColumn        {}
         set select            0
         set state             $options(-state)
         foreach item $column {
            switch -glob -- $item {
               .* {
                  set createRadiobutton 1
                  if {! [regexp {(^[^:]+)(:[nsewc]+$)} $item dummy1 item sticky]} {set sticky {}}
                  set radiobuttonName   $options(name),[string range $item 1 end]
                  if {$options(-group) ne ""} {set gpGroup($radiobuttonName) $options(-group)}  
                  lappend gridColumn $radiobuttonName$sticky
               }
               @* {
                  set gridName .[string range $item 1 end]
                  lappend gridColumn $gridName
               }
               %* {
                  set gpGroup($radiobuttonName) [string range $item 1 end]
               }
               +* {
                  set value [string range $item 1 end]
                  set select 1
               }
               -* {
                  set value [string range $item 1 end]
                  set select 0
               }
               default {
                  lappend gridColumn $item
               }
            }
         }

         if {$createRadiobutton} {
            if {[info exists gpGroup($radiobuttonName)] && [info exists gpGroupState($gpGroup($radiobuttonName))]} {
               set state $gpGroupState($gpGroup($radiobuttonName))
            }
            set ($options(name)) {}
            radiobutton $radiobuttonName -background $options(-background) -disabledforeground black -state $state -takefocus $options(-takefocus) -value $value -variable ($options(name))
            if {$state eq "disabled"} {$radiobuttonName configure -takefocus 0}
            if {$select} {after idle $radiobuttonName select}
         }

         lappend gridData $gridColumn
         incr columnCount
      }
      lappend gridData !!!!
      incr rowCount
   }

   regsub -all {!!!!} $gridData \n gridData 

   set gridCommand "::gridplus::gridplus grid $options(name)"

   foreach option [array names options -*] {
      set gridCommand "$gridCommand $option {$options($option)}"
   }

   set gridCommand "$gridCommand {$gridData}"

   eval $gridCommand
}

#=======================================================================#
# PROC   : ::gridplus::gpSet                                            #
# PURPOSE: Gridplus Set values.                                         #
#=======================================================================#

proc ::gridplus::gpSet {} {
   upvar 1 options options

   variable gpConfig
   variable gpDateFormat
   variable gpGroupState
   variable gpValidation
   variable gpValidationText

   foreach option [array names options -*] {
      switch -- $option {
         -century {
            set gpDateFormat(century) $options(-century)
         }
         -dateformat {
            switch -- $options(-dateformat) {
               eu {
                  set gpDateFormat(day)       0
                  set gpDateFormat(month)     1
                  set gpDateFormat(separator) .
               }
               uk {
                  set gpDateFormat(day)       0
                  set gpDateFormat(month)     1
                  set gpDateFormat(separator) /
               }
               us {
                  set gpDateFormat(day)       1
                  set gpDateFormat(month)     0
                  set gpDateFormat(separator) /
               }
               default {
                  error "GRIDPLUS ERROR: Invalid date format ($options(-dateformat))"
                  return
               }
            }
            set gpConfig(dateformat) $options(-dateformat)
         }
         -errormessage {
            set gpConfig(errormessage) $options(-errormessage)
         }
         -group {
            set gpGroupState($options(-group)) $options(-state)
            ::gridplus::gpSetGroup
         }
         -prefix {
            set gpConfig(prefix) $options(-prefix)
         }
         -proc {
            set gpConfig(proc) $options(-proc)
         }
         -validation {
            if {$options(-pattern) ne ""} {
               set gpValidation($options(-validation)) $options(-pattern)
               if {$options(-text) ne ""} {
                  set gpValidationText($options(-validation)) $options(-text)
               } else {
                  set gpValidationText($options(-validation)) $options(-validation)
               }
            }

         }
      }
   }
}

#=======================================================================#
# PROC   : ::gridplus::gpSetFont                                        #
# PURPOSE: Gridplus Set font attributes.                                #
#=======================================================================#

proc ::gridplus::gpSetFont {font attributes} {

   set fontName [lindex $font 0]
   set fontSize [lindex $font 1]

   set attribute0 [lindex $attributes 0]
   set setFont    1

   switch -regexp -- $attribute0 {
      {^[0-9]+$} {
         set fontSize $attribute0
      } 
      {^[+][0-9]+$} {
         set value [string range $attribute0 1 end]
         set fontSize [expr {$fontSize + $value}]
      }
      {^[-][0-9]+$} {
         set value [string range $attribute0 1 end]
         set fontSize [expr {$fontSize - $value}]
      }
      default {
         set setFont 0
      }
   }

   if {$setFont} {
      return "{$fontName} $fontSize [lrange $attributes 1 end]"
   } else {
      return "$font $attributes"
   }
}

#=======================================================================#
# PROC   : ::gridplus::gpSetFontSize                                    #
# PURPOSE: Gridplus Set font size for "tagged" text widget.             #
#=======================================================================#

proc ::gridplus::gpSetFontSize {defaultSize newSize} {

   switch -regexp -- $newSize {
      {^[0-9]+$} {
         set fontSize $newSize
      } 
      {^[+][0-9]+$} {
         set value [string range $newSize 1 end]
         set fontSize [expr {$defaultSize + $value}]
      }
      {^[-][0-9]+$} {
         set value [string range $newSize 1 end]
         set fontSize [expr {$defaultSize - $value}]
      }
      default {
         set fontSize $defaultSize
      }
   }

   return $fontSize
}

#=======================================================================#
# PROC   : ::gridplus::gpSetGroup                                       #
# PURPOSE: Gridplus Set widgets state to "group" state.                 #
#=======================================================================#

proc ::gridplus::gpSetGroup {} {
   variable gpGroup
   variable gpGroupState

   foreach item [array names gpGroup] {
      if {[info exists gpGroupState($gpGroup($item))]} {
         if {[regexp {^([^@]+)@(.+)$} $item dummy1 configureItem index]} {
            $configureItem entryconfigure $index -state $gpGroupState($gpGroup($item))
         } else {
            $item configure -state $gpGroupState($gpGroup($item))
         }
      }
   }
}

#=======================================================================#
# PROC   : ::gridplus::gpSetStyle                                       #
# PURPOSE: Set options for required style.                              #
#=======================================================================#

proc ::gridplus::gpSetStyle {} {
   upvar 1 options options

   variable gpStyles

   if {$options(-style) eq ""} {return}

   if {[info exists gpStyles($options(-style))]} {
      foreach {option value} $gpStyles($options(-style)) {
         if {$option eq "-pad"} {
            set options(-padx) $value
            set options(-pady) $value
         } else {
            set options($option) $value
         }
      }
   } else {
      error "GRIDPLUS ERROR: Invalid style ($options(-style))"
   }
}

#=======================================================================#
# PROC   : ::gridplus::gpSetTabOrder                                    #
# PURPOSE: Gridplus Set widgets to correct "tab" order.                 #
#=======================================================================#

proc ::gridplus::gpSetTabOrder {name} {
   variable gpTabOrder

   foreach item [lsort [array names gpTabOrder $name:*]] {
      raise $gpTabOrder($item)
      ::gridplus::gpSetTabOrder $gpTabOrder($item)
   }
}

#=======================================================================#
# PROC   : ::gridplus::gpStyle                                          #
# PURPOSE: Create style.                                                #
#=======================================================================#

proc ::gridplus::gpStyle {} {
   upvar 1 options options

   variable gpStyles

   set gpStyles($options(name)) $options(layout)
}

#=======================================================================#
# PROC   : ::gridplus::gpTablelist                                      #
# PURPOSE: Create tablelist.                                            #
#=======================================================================#

proc ::gridplus::gpTablelist {} {
   upvar 1 options options

   global {}

   variable gpGroup
   variable gpGroupState
   variable gpInfo

   set gpInfo($options(name):insertexpr)    $options(-insertexpr)
   set gpInfo($options(name):insertoptions) $options(-insertoptions)
   set gpInfo($options(name):selectfirst)   $options(-selectfirst)
   set gpInfo($options(name):selectmode)    $options(-selectmode)

   set state $options(-state)

   if {$options(-group) ne ""} {
      set gpGroup($options(name).tablelist) $options(-group)
   }
   if {[info exists gpGroup($options(name).tablelist)] && [info exists gpGroupState($gpGroup($options(name).tablelist))]} {
      set state $gpGroupState($gpGroup($options(name).tablelist))
   }

#-------------------------------------#
# Deal with "hide" columns in layout. #
#-------------------------------------#

   set column -1
   set count  0
   set first  0
   set hide   {}
   set index  0

   foreach item $options(layout) {

      if {[string is integer $item]} {
         set  count 0
         incr column
      }
      if {$item eq "hide" && $count == 2} {
         lappend hide $column
         set options(layout) [lreplace $options(layout) $index $index]
         incr index -1
         if {$column == $first} {
            incr first
         }
      }
      incr count
      incr index
   }

   set gpInfo($options(name):selectmode) $options(-selectmode)
   set gpInfo($options(name):sortorder)  $options(-sortorder)

   if {$options(-sortfirst)} {
      set gpInfo($options(name):firstcolumn) 0
   } else {
      set gpInfo($options(name):firstcolumn) $first
   }

   labelframe $options(name) -background $options(-background) -borderwidth $options(-borderwidth) -padx $options(-padx) -pady $options(-pady) -relief $options(-relief) -text [mc $options(-title)]

   tablelist::tablelist $options(name).tablelist                        \
                        -columns         $options(layout)               \
                        -exportselection 0                              \
                        -height          $options(-height)              \
                        -listvariable    $options(-listvariable)        \
                        -selectmode      $options(-selectmode)          \
                        -state           $state                         \
                        -stretch         all                            \
                        -width           $options(-width)               \
                        -xscrollcommand  [list $options(name).xbar set] \
                        -yscrollcommand  [list $options(name).ybar set] \
                        -takefocus       0                              \

   if {$options(-columnsort)} {
      $options(name).tablelist configure -labelcommand ::tablelist::sortByColumn
   }

   scrollbar $options(name).xbar -orient horizontal -command [list $options(name).tablelist xview]
   scrollbar $options(name).ybar -orient vertical   -command [list $options(name).tablelist yview]

   foreach item $hide {
      $options(name).tablelist columnconfigure $item -hide 1
   }

   grid $options(name).tablelist -row 0 -column 0 -sticky news

   switch -- $options(-scroll) {
      x {
         grid $options(name).xbar -row 1 -column 0 -sticky ew
      }
      y {
         grid $options(name).ybar -row 0 -column 1 -sticky ns
      }
      xy {
         grid $options(name).xbar -row 1 -column 0 -sticky ew
         grid $options(name).ybar -row 0 -column 1 -sticky ns
      }
   }

   grid rowconfigure    $options(name) 0 -weight 1
   grid columnconfigure $options(name) 0 -weight 1

   foreach item $options(-options) {
      switch -- $item {
         stripe {
            $options(name).tablelist configure -stripebackground #e0e8f0
         }
         separator {
            $options(name).tablelist configure -showseparators yes
         }
      } 
   }

   if {$options(-proc)} {
      set command "gpProc $options(name)"
   } else {
      set command "$options(-prefix)$options(name)"
      regsub -all {[.]} $command ":" command
      regsub      {^:}  $command {}  command 
   }

   switch -- $options(-action) {
      double {
         bind [$options(name).tablelist bodypath] <Button-1> "after 1 {set ($options(name)) \[::gridplus::gpTablelistSelect $options(name) \[$options(name).tablelist curselection\]\]}"
         bind [$options(name).tablelist bodypath] <Double-1> $command
      }
      single {
         bind [$options(name).tablelist bodypath] <Button-1> "after 1 {set ($options(name)) \[::gridplus::gpTablelistSelect $options(name) \[$options(name).tablelist curselection\]\];$command}"
      }
      default {
         bind [$options(name).tablelist bodypath] <Button-1> "after 1 {set ($options(name)) \[::gridplus::gpTablelistSelect $options(name) \[$options(name).tablelist curselection\]\]}"
      }
   }
}

#=======================================================================#
# PROC   : ::gridplus::gpTablelistInsertExpr                            #
# PURPOSE: Expand tablelist insert expression.                          #
#=======================================================================#

proc ::gridplus::gpTablelistInsertExpr {name line} {
   upvar 1 options options

   variable gpInfo

   if {$gpInfo($name:insertexpr) eq ""} {
      return
   }

   regsub -all -- {%([0-9][0-9]*)} $gpInfo($name:insertexpr) {[lindex $line \1]} insertExpr

   eval "if {$insertExpr} {set result 1} else {set result 0}"

   if {$result} {
      ::gridplus::gpTablelistInsertOptions $name
   }
}

#=======================================================================#
# PROC   : ::gridplus::gpTablelistInsertOptions                         #
# PURPOSE: Process tablelist insert options.                            #
#=======================================================================#

proc ::gridplus::gpTablelistInsertOptions {name} {
   upvar 1 options options

   variable gpInfo

   foreach insertOption $gpInfo($name:insertoptions) {
      if {[lindex $insertOption 0] eq "*"} {
         regsub -- {[*]} $insertOption {end} insertOption
         eval "$name.tablelist rowconfigure $insertOption"
      } else {
         eval "$name.tablelist cellconfigure end,$insertOption"
      }
   }
}

#=======================================================================#
# PROC   : ::gridplus::gpTablelistSelect                                #
# PURPOSE: Sets value for tablelist selections.                         #
#=======================================================================#

proc ::gridplus::gpTablelistSelect {name selection} {
   upvar 1 options options

   global {}

   variable gpInfo

   set count [llength $selection]
   set value [$name.tablelist get $selection]

   if {$gpInfo($name:selectmode) eq "multiple" || $gpInfo($name:selectmode) eq "extended"} {
      if {$count == 1} {
         return [list $value]
      } else {
         return $value
      }
   } else {
      return $value
   }
}

#=======================================================================#
# PROC   : ::gridplus::gpText                                           #
#        : ::gridplus::gpTextSet                                        #
# PURPOSE: Create text.                                                 #
#=======================================================================#

proc ::gridplus::gpText {} {
   upvar 1 options options

   global {}

   variable gpGroup
   variable gpGroupState
   variable gpInfo

   set state $options(-state)

   if {$options(-group) ne ""} {
      set gpGroup($options(name).text) $options(-group)
   }
   if {[info exists gpGroup($options(name).text)] && [info exists gpGroupState($gpGroup($options(name).text))]} {
      set state $gpGroupState($gpGroup($options(name).text))
   }

   labelframe $options(name) -background $options(-background) -borderwidth $options(-borderwidth) -padx $options(-padx) -pady $options(-pady) -relief $options(-relief) -text [mc $options(-title)]

   text $options(name).text                                                                           \
        -background     white                                                                         \
        -height         $options(-height)                                                             \
        -state          $state                                                                        \
        -tabs           {0.5c 1c 1.5c 2c 2.5c 3.0c 3.5c 4.0c 4.5c 5.0c 5.5c 6.0c 6.5c 7.0c 7.5c 8.0c} \
        -takefocus      $options(-takefocus)                                                          \
        -width          $options(-width)                                                              \
        -wrap           $options(-wrap)                                                               \
        -xscrollcommand [list $options(name).xbar set]                                                \
        -yscrollcommand [list $options(name).ybar set]                                                \

   scrollbar $options(name).xbar -orient horizontal -command [list $options(name).text xview]
   scrollbar $options(name).ybar -orient vertical   -command [list $options(name).text yview]

   grid $options(name).text -row 0 -column 0 -sticky news

   switch -- $options(-scroll) {
      x {
         grid $options(name).xbar -row 1 -column 0 -sticky ew
      }
      y {
         grid $options(name).ybar -row 0 -column 1 -sticky ns
      }
      xy {
         grid $options(name).xbar -row 1 -column 0 -sticky ew
         grid $options(name).ybar -row 0 -column 1 -sticky ns
      }
   }

   grid rowconfigure    $options(name) 0 -weight 1
   grid columnconfigure $options(name) 0 -weight 1

   if {$options(-tags)} {
      set normalColor [lindex [split $options(-linkcolor) /] 0]
      set overColor   [lindex [split $options(-linkcolor) /] 1]
      set normalStyle [lindex [split $options(-linkstyle) /] 0]
      set overStyle   [lindex [split $options(-linkstyle) /] 1]

      regsub -- {[&]} $overStyle $normalStyle, overStyle

      if {! [string match */* $options(-linkcolor)]} {set overColor $normalColor}
      if {! [string match */* $options(-linkstyle)]} {set overStyle $normalStyle}

      if {$normalColor eq ""} {set normalColor "blue"}
      if {$overColor   eq ""} {set overColor "blue"}

      if {$normalStyle eq "underline"} {
         set normalStyle "true"
      } else {
         set normalStyle "false"
      }
      if {$overStyle eq "underline"} {
         set overStyle "true"
      } else {
         set overStyle "false"
      }

      set gpInfo($options(name):bgcolor)     white
      set gpInfo($options(name):defaultbg)   white
      set gpInfo($options(name):defaultfg)   black
      set gpInfo($options(name):defaultfont) helvetica
      set gpInfo($options(name):defaultsize) [lindex [$options(name).text cget -font] 1]
      set gpInfo($options(name):fgcolor)     black
      set gpInfo($options(name):font)        [lindex [$options(name).text cget -font] 0]
      set gpInfo($options(name):iconlibrary) [file join $options(-iconpath) $options(-iconfile)]
      set gpInfo($options(name):indent)      0
      set gpInfo($options(name):link)        blue
      set gpInfo($options(name):linkcursor)  $options(-linkcursor)
      set gpInfo($options(name):normalcolor) $normalColor
      set gpInfo($options(name):normalstyle) $normalStyle
      set gpInfo($options(name):overcolor)   $overColor
      set gpInfo($options(name):overstyle)   $overStyle
      set gpInfo($options(name):prefix)      $options(-prefix)
      set gpInfo($options(name):proc)        $options(-proc)
      set gpInfo($options(name):size)        [lindex [$options(name).text cget -font] 1]
      set gpInfo($options(name):tagid)       0
      set gpInfo($options(name):tags)        1

      $options(name).text configure -cursor {} -state disabled
   } else {
      set gpInfo($options(name):tags)        0
   }

   bind $options(name).text <KeyRelease> "::gridplus::gpTextSet $options(name)"
   bind $options(name).text <Tab>        "[bind all <Tab>];break"
   bind $options(name).text <Shift-Tab>  "[bind all <<PrevWindow>>]; break"

   set ($options(name)) {}
}

#=======================================================================#
# PROC   : ::gridplus::gpTextSet                                        #
# PURPOSE: Set contents of GRIDPLUS Text.                               #
#=======================================================================#

proc ::gridplus::gpTextSet {item} {
   global {}

   if {[$item.text edit modified]} {
      set ($item) {}

      foreach {key text index} [$item.text dump -text 1.0 end] {
         set ($item) "$($item)$text"
      }

      $item.text edit modified 0
   }
}

#=======================================================================#
# PROC   : ::gridplus::gpTree                                           #
# PURPOSE: Create tree.                                                 #
#=======================================================================#

proc ::gridplus::gpTree {} {
   upvar 1 options options

   global {}

   variable gpInfo

   set gpInfo($options(name):fileicon)     $options(-fileicon)
   set gpInfo($options(name):foldericon)   $options(-foldericon)
   set gpInfo($options(name):iconfile)     $options(-iconfile)
   set gpInfo($options(name):iconpath)     $options(-iconpath)
   set gpInfo($options(name):treemode)     $options(-treemode)

   labelframe $options(name) -background $options(-background) -borderwidth $options(-borderwidth) -padx $options(-padx) -pady $options(-pady) -relief $options(-relief) -text [mc $options(-title)]

   text $options(name).tree                            \
        -background     white                          \
        -cursor         left_ptr                       \
        -height         $options(-height)              \
        -spacing3       2                              \
        -state          disabled                       \
        -tabs           20                             \
        -width          $options(-width)               \
        -wrap           none                           \
        -xscrollcommand [list $options(name).xbar set] \
        -yscrollcommand [list $options(name).ybar set]

   bindtags $options(name).tree [list $options(name).tree . all]

   $options(name).tree tag configure selection -background [$options(name).tree cget -selectbackground] -foreground white

   scrollbar $options(name).xbar -orient horizontal -command [list $options(name).tree xview]
   scrollbar $options(name).ybar -orient vertical   -command [list $options(name).tree yview]

   grid $options(name).tree -row 0 -column 0 -sticky news

   switch -- $options(-scroll) {
      x {
         grid $options(name).xbar -row 1 -column 0 -sticky ew
      }
      y {
         grid $options(name).ybar -row 0 -column 1 -sticky ns
      }
      xy {
         grid $options(name).xbar -row 1 -column 0 -sticky ew
         grid $options(name).ybar -row 0 -column 1 -sticky ns
      }
   }

   grid rowconfigure    $options(name) 0 -weight 1
   grid columnconfigure $options(name) 0 -weight 1

   if {$options(-proc)} {
      set command "gpProc $options(name)"
   } else {
      set command "$options(-prefix)$options(name)"
      regsub -all {[.]} $command ":" command
      regsub      {^:}  $command {}  command 
   }

   switch -- $options(-action) {
      double {
         bind $options(name).tree <Button-1> "after 1 ::gridplus::gpTreeSelect $options(name) @%x,%y"
         bind $options(name).tree <Double-1> "after 1 $command"
      }
      single {
         bind $options(name).tree <Button-1> "after 1 ::gridplus::gpTreeSelect $options(name) @%x,%y $command"
      }
      default {
         bind $options(name).tree <Button-1> "after 1 ::gridplus::gpTreeSelect $options(name) @%x,%y"
      }
   }

   if {[lsearch [image names] ::icon::$options(-fileicon)] < 0} {
      ::icons::icons create -file [file join $options(-iconpath) $options(-iconfile)] $options(-fileicon)
   }
   if {[lsearch [image names] ::icon::$options(-foldericon)] < 0} {
      ::icons::icons create -file [file join $options(-iconpath) $options(-iconfile)] $options(-foldericon)
   }

   set ($options(name)) {}
}

#=======================================================================#
# PROC   : ::gridplus::gpTreeContract                                   #
# PURPOSE: Contract branch of GRIDPLUS Tree.                            #
#=======================================================================#

proc ::gridplus::gpTreeContract {name index path} {

   variable gpInfo

   $name.tree image configure $index -name contract:$path -image ::icon::expand

   foreach tag [$name.tree tag names] {
      if {[string match path:$path/* $tag]} {$name.tree tag configure $tag -elide 1}
   }
}

#=======================================================================#
# PROC   : ::gridplus::gpTreeExpand                                     #
# PURPOSE: Expand branch of GRIDPLUS Tree.                              #
#=======================================================================#

proc ::gridplus::gpTreeExpand {name index path} {

   foreach tag [$name.tree tag names] {
      if {[string match path:$path/ $tag]} {$name.tree tag configure $tag -elide 0}
   }

   $name.tree image configure $index -name expand:$path -image ::icon::contract

   foreach image [$name.tree image names] {
      regexp {^([^:]+):(.*)$} [$name.tree image cget $image -name] dummy imageType imagePath
      if {[regexp "^$path/(.+)$" $imagePath dummy imageBasename]} {
         if {[llength [split $imageBasename /]] == 1 && "$imageType" eq "expand"} {
           ::gridplus::gpTreeExpand $name $image $imagePath
         }
      }
   } 
}

#=======================================================================#
# PROC   : ::gridplus::gpTreeSelect                                     #
# PURPOSE: Sets value for tree selections.                              #
#=======================================================================#

proc ::gridplus::gpTreeSelect {name index {command {}}} {

   global {}

   if {! [catch {set icon [$name.tree image cget $index -name]}]} {
      regexp {^([^:]+):(.+)$} $icon dummy type path
      switch -- $type {
         contract {::gridplus::gpTreeExpand   $name $index $path; return}
         expand   {::gridplus::gpTreeContract $name $index $path; return}
      }
   }

   set space [$name.tree search -regexp { } "$index linestart" "$index lineend"]

   if {$space eq ""} {return}

   foreach tag [$name.tree tag names $index] {
      if {[string match item:* $tag]} {regsub -all "\034" [string range $tag 5 end] { } ($name)}
   }

   $name.tree tag remove selection 1.0 end
   $name.tree tag add selection "$space + 1 chars" "$index lineend"

   if {$command ne ""} {$command}
}

#=======================================================================#
# PROC   : ::gridplus::gpTreeSet                                        #
# PURPOSE: Set contents of GRIDPLUS Tree.                               #
#=======================================================================#

proc ::gridplus::gpTreeSet {name nodes} {

   variable gpInfo

   $name.tree configure -state normal
   $name.tree delete 1.0 end

   set line 1

   foreach node $nodes {
      set icon     {}
      set nodeText {}
      set nodeType file

      foreach item $node {
         switch -regexp -- $item {
            ^: {
               set icon [string range $item 1 end]
            }
            ^[+]$ {
               set nodeType folder
            }
            ^[/] {
               regsub -all { } $item "\034" nodeFullName
            }
            default {
               set nodeText $item
            }
         }

      }

      if {[regexp {^(.*/)([^/]+)$} $nodeFullName dummy path nodeName]} {
         set indent [string repeat "\t" [expr {[llength [split $path "/"]] - 2}]]
      } else {
         set path     $nodeFullName
         set nodeName $nodeFullName
         set indent   ""
      }

      if {$nodeText ne ""} {
         set nodeName $nodeText
      } else {
         regsub -all "\034" $nodeName { } nodeName
      }

      set nodeName [mc $nodeName]

      if {$icon eq ""} {
         set icon $gpInfo($name:${nodeType}icon)
      } else {
         if {[lsearch [image names] ::icon::$icon] < 0} {
            ::icons::icons create -file [file join $gpInfo($name:iconpath) $gpInfo($name:iconfile)] $icon
         }
      }

      if {$nodeType eq "folder"} {
         if {$gpInfo($name:treemode) eq "expand"} {
            set actionIcon contract
         } else {
            set actionIcon expand
         }
      } else {
         set actionIcon spacer
      }

      if {[llength [split $path "/"]] == 2} {
         $name.tree tag configure path:$path -elide 0
      } else {
         if {$gpInfo($name:treemode) eq "contract"} {
            $name.tree tag configure path:$path -elide 1
         } else {
            $name.tree tag configure path:$path -elide 0
         }
      }

      $name.tree configure -state normal

      $name.tree insert end $indent
      $name.tree image  create end -image ::icon::$actionIcon -name "action:$nodeFullName" -padx 4
      $name.tree image  create end -image ::icon::$icon -name "icon:$line"
      $name.tree insert end " $nodeName\n"
      $name.tree tag    add path:$path $line.0 "$line.end + 1 chars"
      $name.tree tag    add item:$nodeFullName $line.0 "$line.end + 1 chars"

      switch -- $actionIcon {
         contract {$name.tree image configure action:$nodeFullName -name expand:$nodeFullName}
         expand   {$name.tree image configure action:$nodeFullName -name contract:$nodeFullName}
      }

      $name.tree configure -state disabled

      incr line
   }
}

#=======================================================================#
# PROC   : ::gridplus::gpValidateFailed                                 #
# PURPOSE: Sets focus to failed validation entry.                       #
#=======================================================================#

proc ::gridplus::gpValidateFailed {widget} {

   variable gpInfo

   set focus [focus]

   if {[string compare {} $focus] && [winfo class $focus] eq "Entry"} {
      $focus selection clear
      if {[regexp {^(focus(out)?|all)} [set validate [$focus cget -validate]]]} {
         $focus configure -validate none
         after idle [list $focus configure -validate $validate]
      }
   }
   after 1 [list focus $widget]
}

#=======================================================================#
# PROC   : ::gridplus::gpValidate                                       #
# PURPOSE: Validates contents of entry.                                 #
#=======================================================================#

proc ::gridplus::gpValidate {entry validation} {
   global {}

   variable gpConfig
   variable gpInfo
   variable gpValidation
   variable gpValidationText

   if {[info exists gpInfo(validation:failed)] && $gpInfo(validation:failed) ne $entry} {
      return 1
   }

   set focus [focus]

   if {$focus eq "" || [winfo class $focus] eq "Toplevel" || [winfo class $focus] eq "Button"} {
      return 1
   }

   if {! [regexp {^([.][^.,]+)} $entry dummy window]} {
      set window {}
   } else {
      if {[winfo class $window] ne "Toplevel"} {
         set window {}
      }
   }

   set validationOK 0

   if {! [regexp -- {([^:]+):(.*)} $validation dummy validation parameter]} {
      set parameter {}
   }

   switch -glob -- $gpValidation($validation) {
      proc:* {
         set validateProc [string range $gpValidation($validation) 5 end]
         if {[$validateProc $entry $parameter]} {
            set validationOK 1
         }
      }
      default {
         if {[regexp $gpValidation($validation) $($entry)]} {
            set validationOK 1
         }
      }
   }

   if $validationOK {
      $entry configure -background white
      $entry configure -foreground black

      if {[winfo exists $window.errormessage]} {
         $window.errormessage configure -text {}
      }

      unset -nocomplain gpInfo(validation:failed)

      return 1
   } else {

# Is the following IF still required???

      if {[info exists gpConfig(errormessage)]} {
         set errorText    [mc $gpValidationText($validation)]
         set errorMessage [mc $gpConfig(errormessage)]
         regsub {%} $errorText    $parameter errorText
         regsub {%} $errorMessage $errorText errorMessage
      }

      if {[winfo exists $window.errormessage]} {
         $window.errormessage configure -text $errorMessage
      }

      $entry configure -background red
      $entry configure -foreground white

      set gpInfo(validation:failed) $entry

      return 0
   }
}

#=======================================================================#
# PROC   : ::gridplus::gpWindow                                         #
# PURPOSE: Create toplevel window with "modal" option.                  #
#=======================================================================#

proc ::gridplus::gpWindow {} {
   upvar 1 options options

   if {[winfo exists $options(name)]} {
      if {$options(-deletewindow) ne ""} {
         wm protocol $options(name) WM_DELETE_WINDOW "after 1 {$options(-deletewindow)}"
      }
      return 0
   }

   toplevel $options(name)

   wm resizable $options(name) 0 0

   bind $options(name) <Destroy> "::gridplus::gpButtonHelpCancel"

   if {$options(-deletewindow) ne ""} {
      wm protocol $options(name) WM_DELETE_WINDOW "after 1 {$options(-deletewindow)}"
   }

   if {$options(-modal)} {
      bind modalWindow <ButtonPress> {wm deiconify %W;raise %W}
      bindtags $options(name) [linsert [bindtags $options(name)] 0 modalWindow]

      wm deiconify $options(name)
      tkwait visibility $options(name)
      grab set $options(name)
   }

   return 1
}

#=======================================================================#
# PROC   : ::gridplus::gpValidateDate                                   #
# PURPOSE: Validates for valid date.                                    #
#=======================================================================#

proc ::gridplus::gpValidateDate {entry parameter} {
   global {}

   if {[catch {clock scan [gpFormatDate $($entry) internal]}] == 0} {
      set ($entry) [gpFormatDate $($entry) application]
      $entry configure -validate focusout
      return 1
   } else {
      return 0
   }
}

#=======================================================================#
# PROC   : ::gridplus::gpFormatDate                                     #
# PURPOSE: Converts date to standard (US) format for validation.        #
#=======================================================================#

proc ::gridplus::gpFormatDate {date mode} {
   variable gpDateFormat

   switch -regexp -- $date {
      {^[0-9]{6}$} {
         set part(0) [string range $date 0 1]
         set part(1) [string range $date 2 3]
         set year    [string range $date 4 5]
         if {$year <= $gpDateFormat(century)} {
            set year "20$year"
         } else {
            set year "19$year"
         }
      }
      {^[0-9]{8}$} {
         set part(0) [string range $date 0 1]
         set part(1) [string range $date 2 3]
         set year    [string range $date 4 7]
      }
      {^[0-9]{2}.[0-9]{2}.[0-9]{4}$}  {
         set part(0) [string range $date 0 1]
         set part(1) [string range $date 3 4]
         set year    [string range $date 6 9]
      }
      default  {
         set part(0) x
         set part(1) x
         set year    x
      }
   }

   set separator $gpDateFormat(separator)

   if {[string equal $mode internal]} {
      return $part($gpDateFormat(month))/$part($gpDateFormat(day))/$year
   } else {
      return $part(0)$separator$part(1)$separator$year
   }
}

#=======================================================================#
# PROC   : ::gridplus::gpmap                                            #
# PURPOSE: Map Gridplus "variable(s)" to a list of values.              #
#=======================================================================#

proc ::gridplus::gpmap {map values} {

   set item 0

   foreach value $values {
      gpset [lindex $map $item] $value
      incr item
   }
}

#=======================================================================#
# PROC   : ::gridplus::gpset                                            #
# PURPOSE: Set Gridplus "variable(s)".                                  #
#=======================================================================#

proc ::gridplus::gpset {args} {
   global {}

   variable gpInfo

   switch -- [llength $args] {
      1 {
         if {[expr [llength [lindex $args 0]] % 2] != 0} {
            error "GRIDPLUS ERROR: (gpset) Unmatched item/value"
         }
         foreach {item value} [lindex $args 0] {
            if {[winfo exists $item.text]} {
               $item.text delete 1.0 end
               $item.text insert end $value
               set ($item) $value
            } else {
               set ($item) $value
            }
         }
      }
      2 {
         set item  [lindex $args 0]
         set value [lindex $args 1]
         if {[winfo exists $item.tablelist]} {
            unset -nocomplain ($item)
            $item.tablelist delete 0 end
            foreach line $value {
               $item.tablelist insert end $line
               gpTablelistInsertExpr $item $line
            }
            $item.tablelist sortbycolumn $gpInfo($item:firstcolumn) -$gpInfo($item:sortorder)
            if {$gpInfo($item:selectfirst)} {
               $item.tablelist selection set 0
               set ($item) [$item.tablelist get 0]
            }
         } elseif {[winfo exists $item.text]} {
            set textState [$item.text cget -state]
            $item.text configure -state normal
            if {$gpInfo($item:tags)} {
               $item.text delete 1.0 end
               ::gridplus::gpParseTags $item $value end
            } else {
               $item.text delete 1.0 end
               $item.text insert end $value
               set ($item) $value
            }
            $item.text configure -state $textState
         } elseif {[winfo exists $item.tree]} {
            ::gridplus::gpTreeSet $item $value
         } else {
            set ($item) $value
         }
      }
      default {
         error "GRIDPLUS ERROR: (gpset) Wrong number of Args"
      }
   }
}

#=======================================================================#
# PROC   : ::gridplus::gpunset                                          #
# PURPOSE: Unset Gridplus "variable(s)".                                #
#=======================================================================#

proc ::gridplus::gpunset {args} {
   global {}

   foreach arg $args {
      if {[info exists ($arg)]} {
         unset ($arg)
      }
      if {[winfo exists $arg.tablelist]} {
         $arg.tablelist delete 0 end
      } elseif {[winfo exists $arg.text]} {
         $arg.text delete 1.0 end
      } elseif {[winfo exists $arg.tree]} {
         $arg.tree configure -state normal
         $arg.tree delete 1.0 end
         $arg.tree configure -state disabled
      }
   }
}

#=======================================================================#
# End of Script: gridplus.tcl                                           #
#=======================================================================#
